use std::cell::RefCell;
use crate::ast::{Identifier, UseStatement};
use std::fmt::Display;
use std::range::Range;
use std::rc::Rc;

#[derive(Clone, Debug)]
pub struct SourceDefinition {
    file_id: usize,
    range: Range<usize>,
}

impl SourceDefinition {
    pub fn from_identifier(identifier: &Identifier) -> Self {
        SourceDefinition {
            file_id: identifier.file_id,
            range: identifier.range,
        }
    }
    
    pub fn from_identifier_rc(identifier: Rc<RefCell<Identifier>>) -> Self {
        let borrowed = identifier.borrow();
        SourceDefinition {
            file_id: borrowed.file_id,
            range: borrowed.range,
        }
    }

    #[deprecated(note = "Use identifier instead.")]
    pub fn from_use_statement(use_statement: &UseStatement) -> Self {
        SourceDefinition {
            file_id: use_statement.file_id,
            range: use_statement.range,
        }
    }

    pub fn file_id(&self) -> usize {
        self.file_id
    }

    pub fn range(&self) -> Range<usize> {
        self.range.clone()
    }
}

pub trait SymbolInner {
    fn declared_name(&self) -> &str;
    fn definition(&self) -> Option<SourceDefinition>;
}

/* Symbol */

#[derive(Debug, Clone)]
pub enum Symbol {
    UseStatement(Rc<RefCell<UseStatementSymbol>>),
    Module(Rc<ModuleSymbol>),
    Type(Rc<TypeSymbol>),
    Function(Rc<FunctionSymbol>),
    Parameter(Rc<ParameterSymbol>),
    Variable(Rc<VariableSymbol>),
}

impl Symbol {
    pub fn definition(&self) -> Option<SourceDefinition> {
        match self {
            Symbol::UseStatement(s) => s.borrow().definition.clone(),
            Symbol::Module(s) => s.definition(),
            Symbol::Type(s) => s.definition(),
            Symbol::Function(s) => s.definition(),
            Symbol::Parameter(s) => s.definition(),
            Symbol::Variable(s) => s.definition(),
        }
    }
    
    pub fn unwrap_use_statement_symbol(&self) -> Rc<RefCell<UseStatementSymbol>> {
        match self {
            Symbol::UseStatement(s) => s.clone(),
            _ => panic!("unwrap_use_statement_symbol called on non-use statement symbol"),
        }
    }
}

impl Display for Symbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        use Symbol::*;
        match self {
            UseStatement(use_statement_symbol) => write!(f, "{}", use_statement_symbol.borrow()),
            Module(module_symbol) => write!(f, "{}", module_symbol),
            Type(class_symbol) => write!(f, "{}", class_symbol),
            Function(function_symbol) => write!(f, "{}", function_symbol),
            Parameter(parameter_symbol) => write!(f, "{}", parameter_symbol),
            Variable(variable_symbol) => write!(f, "{}", variable_symbol),
        }
    }
}

/* Use-statement */

#[derive(Debug, Clone)]
pub struct UseStatementSymbol {
    pub fqn: String,
    pub declared_name: String,
    definition: Option<SourceDefinition>,
    referenced_symbol: Option<Box<Symbol>>
}

impl UseStatementSymbol {
    pub fn new(fqn: &str, declared_name: &str, identifier: Option<Rc<RefCell<Identifier>>>) -> Self {
        UseStatementSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            definition: identifier.map(SourceDefinition::from_identifier_rc),
            referenced_symbol: None,
        }
    }

    pub fn set_referenced_symbol(&mut self, referenced_symbol: Symbol) {
        self.referenced_symbol = Some(Box::new(referenced_symbol));
    }

    pub fn referenced_symbol(&self) -> Option<Box<Symbol>> {
        self.referenced_symbol.clone()
    }
}

impl SymbolInner for UseStatementSymbol {
    fn declared_name(&self) -> &str {
        &self.declared_name
    }

    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for UseStatementSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "UseStatementSymbol(fqn = {}, declared_name = {})",
            self.fqn, self.declared_name
        )
    }
}


/* Module */

#[derive(Debug)]
pub struct ModuleSymbol {
    fqn: String,
    declared_name: String,
    is_public: bool,
    definition: Option<SourceDefinition>,
}

impl ModuleSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        identifier: Option<&Identifier>,
    ) -> ModuleSymbol {
        ModuleSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: identifier.map(SourceDefinition::from_identifier),
        }
    }
}

impl SymbolInner for ModuleSymbol {
    fn declared_name(&self) -> &str {
        self.declared_name.as_str()
    }

    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for ModuleSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "ModuleSymbol(name = {}, is_public = {})",
            self.fqn, self.is_public
        )
    }
}

/* Class */

#[derive(Debug)]
pub struct TypeSymbol {
    fqn: String,
    declared_name: String,
    is_public: bool,
    definition: Option<SourceDefinition>,
}

impl TypeSymbol {
    pub fn new(fqn: &str, declared_name: &str, is_public: bool, identifier: Option<&Identifier>) -> Self {
        TypeSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            definition: identifier.map(SourceDefinition::from_identifier),
        }
    }

    pub fn fqn(&self) -> &str {
        &self.fqn
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }
}

impl SymbolInner for TypeSymbol {
    fn declared_name(&self) -> &str {
        &self.declared_name
    }

    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for TypeSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "TypeSymbol(fqn = {}, declared_name = {})",
            self.fqn, self.declared_name
        )
    }
}

/* Function */

#[derive(Debug)]
pub struct FunctionSymbol {
    fqn: String,
    declared_name: String,
    is_public: bool,
    is_platform: bool,
    definition: Option<SourceDefinition>,
}

impl FunctionSymbol {
    pub fn new(
        fqn: &str,
        declared_name: &str,
        is_public: bool,
        is_platform: bool,
        identifier: Option<&Identifier>,
    ) -> FunctionSymbol {
        FunctionSymbol {
            fqn: fqn.to_string(),
            declared_name: declared_name.to_string(),
            is_public,
            is_platform,
            definition: identifier.map(SourceDefinition::from_identifier),
        }
    }
    
    pub fn fqn(&self) -> &str {
        &self.fqn
    }
}

impl SymbolInner for FunctionSymbol {
    fn declared_name(&self) -> &str {
        self.declared_name.as_str()
    }
    
    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for FunctionSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "FunctionSymbol(fqn = {}, declared_name = {}, is_public = {})",
            self.fqn, self.declared_name, self.is_public
        )
    }
}

/* Parameter */

#[derive(Debug)]
pub struct ParameterSymbol {
    declared_name: String,
    definition: Option<SourceDefinition>,
}

impl ParameterSymbol {
    pub fn new(declared_name: &str, identifier: Option<&Identifier>) -> Self {
        ParameterSymbol {
            declared_name: declared_name.to_string(),
            definition: identifier.map(SourceDefinition::from_identifier),
        }
    }
}

impl SymbolInner for ParameterSymbol {
    fn declared_name(&self) -> &str {
        &self.declared_name
    }
    
    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for ParameterSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "ParameterSymbol({})",
            self.declared_name
        )
    }
}

/* Variable */

#[derive(Debug)]
pub struct VariableSymbol {
    declared_name: String,
    is_mutable: bool,
    definition: Option<SourceDefinition>,
}

impl VariableSymbol {
    pub fn new(declared_name: &str, is_mutable: bool, identifier: Option<&Identifier>) -> Self {
        VariableSymbol {
            declared_name: declared_name.to_string(),
            is_mutable,
            definition: identifier.map(SourceDefinition::from_identifier),
        }
    }
}

impl SymbolInner for VariableSymbol {
    fn declared_name(&self) -> &str {
        self.declared_name.as_str()
    }
    
    fn definition(&self) -> Option<SourceDefinition> {
        self.definition.clone()
    }
}

impl Display for VariableSymbol {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(
            f,
            "VariableSymbol(name = {}, is_mutable = {})",
            self.declared_name, self.is_mutable
        )
    }
}
